#!/usr/bin/env python3

import concurrent.futures, configparser, pathlib, sys, subprocess, traceback

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('GtkSource', '3.0')
from gi.repository import Gtk, Gdk, GtkSource, Pango

from zehkira_mili import Defaults, Paths


class App:
	# buffer modified-changed signal
	def _on_text_modified(self, *_):
		indicator = ' *' if self._view.get_buffer().get_modified() else ''
		self._barModified.set_label(indicator)
		self._window.set_title('Mili - ' + self._file + indicator)

	# fire functions based on keyboard event
	def _on_key_event(self, _, event):
		modType = Gdk.ModifierType
		validMask = (
			modType.SHIFT_MASK
			| modType.HYPER_MASK
			| modType.MOD2_MASK
			| modType.MOD3_MASK
			| modType.CONTROL_MASK
			| modType.MOD4_MASK
			| modType.SUPER_MASK
			| modType.MOD1_MASK
		)
		key = (event.keyval, event.state & validMask)
		if key in self._keybinds:
			self._do_safely(self._keybinds[key])

	# clean app exit
	def _on_exit(self, *_) -> bool:
		# return True to NOT exit (due to GTK event handling)
		return self._popup_ask_save() == Gtk.ResponseType.CANCEL

	# constructor, duh.
	def __init__(self):
		# window and events
		self._window = Gtk.ApplicationWindow(title = 'Mili')
		self._window.add_events(Gdk.EventMask.KEY_PRESS_MASK)
		self._window.connect('delete_event', self._on_exit)
		self._window.connect('key_press_event', self._on_key_event)

		# icon
		self._window.set_icon_from_file(Paths.icon)

		# set up scrolling container
		scroller = Gtk.ScrolledWindow(hadjustment = None, vadjustment = None)
		self._view = GtkSource.View()
		self._view.set_monospace(True)
		self._view.get_buffer().connect(
			'modified_changed',
			self._on_text_modified
		)
		scroller.add(self._view)

		# status bar labels
		self._barFile = Gtk.Label(label = '')
		self._barFile.set_single_line_mode(True)
		self._barFile.set_halign(Gtk.Align.START)
		self._barModified = Gtk.Label(label = '')
		self._barModified.set_single_line_mode(True)
		self._barModified.set_halign(Gtk.Align.START)
		self._barIndent = Gtk.Label(label = '')
		self._barIndent.set_halign(Gtk.Align.END)
		self._barIndent.set_single_line_mode(True)

		# status (bottom) bar
		botBar = Gtk.Box(orientation = Gtk.Orientation.HORIZONTAL, spacing = 0)
		botBar.set_homogeneous(False)
		botBar.pack_start(self._barFile, False, False, 0)
		botBar.pack_start(self._barModified, False, False, 0)
		botBar.pack_end(self._barIndent, False, False, 0)

		# find and replace interface
		self._entryFind = Gtk.Entry()
		self._entryReplace = Gtk.Entry()
		self._entryFind.set_placeholder_text('Find...')
		self._entryReplace.set_placeholder_text('Replace with...')
		find = Gtk.Button.new_with_label('Find next')
		replace = Gtk.Button.new_with_label('Replace this')
		find.connect('clicked', self._find)
		replace.connect('clicked', self._replace)

		# top bar for find and replace
		self._finder = Gtk.Box(
			orientation = Gtk.Orientation.HORIZONTAL,
			spacing = 0
		)
		self._finder.set_homogeneous(False)
		self._finder.pack_start(self._entryFind, True, True, 0)
		self._finder.pack_start(self._entryReplace, True, True, 0)
		self._finder.pack_start(find, False, False, 0)
		self._finder.pack_start(replace, False, False, 0)

		# pack everything into top level container
		box = Gtk.Box(orientation = Gtk.Orientation.VERTICAL, spacing = 0)
		box.set_homogeneous(False)
		box.pack_start(self._finder, False, False, 0)
		box.pack_start(scroller, True, True, 0)
		box.pack_start(botBar, False, False, 0)

		self._window.add(box)

		# app data
		self._file = ''
		self._smartIndent = False
		self._keybinds = {}

		# config
		self._do_safely(self._configure)

		# this MUST happen BEFORE the next step!
		self._window.show_all()
		self._finder.set_visible(False)
		self._view.grab_focus()
		self._window.connect('destroy', Gtk.main_quit)

		# pass file as argument
		if len(sys.argv) > 1:
			self._do_safely(self._open, sys.argv[1])

		Gtk.main()

	# do something while catching errors and preventing text editing
	def _do_safely(self, action: callable, data: any = None) -> any:
		self._view.set_editable(False)
		result = None

		# handle an argument, no arguments, and any return value
		try:
			if data:
				result = action(data)
			else:
				result = action()
		except:
			err = str(action) +': ' + traceback.format_exc()
			print(err)
			self._popup_error(err)

		self._view.set_editable(True)
		return result

	# applies configs in [default -> global -> local] order
	def _configure(self):
		# apply the default config
		config = configparser.ConfigParser()
		config.read_string(Defaults.config)
		self._apply_config(config)

		# apply the global and local configs
		for path in [Paths.globalConf, Paths.homeConf]:
			pathObj = pathlib.Path(path)
			if pathObj.exists():
				config = configparser.ConfigParser()
				config.read(path)
				self._apply_config(config)

	# applies the given conf, ignoring missing keys and sections
	def _apply_config(self, config: configparser.ConfigParser):
		bfr = self._view.get_buffer()

		if 'Window' in config:
			window = config['Window']
			width, height = self._window.get_default_size()
			width = window.getint('Width') if 'Width' in window else width
			height = window.getint('Height') if 'Height' in window else height
			self._window.set_default_size(width, height)

		if 'Bar' in config:
			bar = config['Bar']
			if 'File' in bar:
				self._barFile.set_visible(bar.getboolean('File'))
			if 'Modified' in bar:
				self._barModified.set_visible(bar.getboolean('Modified'))
			if 'Indent' in bar:
				self._barIndent.set_visible(bar.getboolean('Indent'))

		if 'Indent' in config:
			indent = config['Indent']
			if 'Auto' in indent:
				self._view.set_auto_indent(indent.getboolean('Auto'))
			if 'Size' in indent:
				self._view.set_indent_width(indent.getint('Size'))
				self._view.set_tab_width(indent.getint('Size'))
			if 'Prefer' in indent:
				self._view.set_insert_spaces_instead_of_tabs(
					indent['Prefer'] == 'spaces'
				)
			if 'Tablike' in indent:
				self._view.set_smart_backspace(indent.getboolean('Tablike'))
			if 'Smart' in indent:
				self._smartIndent = indent.getboolean('Smart')

		# need to reset to default here, so update status bar
		if self._view.get_insert_spaces_instead_of_tabs():
			label = str(self._view.get_indent_width()) + ' spaces '
		else:
			label = 'tabs '
		self._barIndent.set_label(label)

		if 'Whitespace' in config:
			white = config['Whitespace']
			drawer = self._view.get_space_drawer()

			# expected keyword in config : type flag
			types = {
				'spaces':  GtkSource.SpaceTypeFlags.SPACE,
				'tabs':  GtkSource.SpaceTypeFlags.TAB,
				'nonbreaking':  GtkSource.SpaceTypeFlags.NBSP,
				'newlines':  GtkSource.SpaceTypeFlags.NEWLINE,
			}

			# expected option in Whitespace section : location flag
			locs = {
				'Leading': GtkSource.SpaceLocationFlags.LEADING,
				'Trailing': GtkSource.SpaceLocationFlags.TRAILING,
				'Internal': GtkSource.SpaceLocationFlags.INSIDE_TEXT
			}

			# set flags for locations - that includes disabling
			for _, (setting, loc) in enumerate(locs.items()):
				# it's ok to only set for some of the locations
				if setting in white:
					typesForLoc = GtkSource.SpaceTypeFlags.NONE
					for _, (keyword, typ) in enumerate(types.items()):
						if keyword in white[setting].lower():
							typesForLoc |= typ

					# will disable ws for loc if no valid keywords are found
					drawer.set_types_for_locations(loc, typesForLoc)

			drawer.set_enable_matrix(True)

		if 'Margin' in config:
			margin = config['Margin']
			if 'Show' in margin:
				self._view.set_show_right_margin(margin.getboolean('Show'))
			if 'Position' in margin:
				self._view.set_right_margin_position(margin.getint('Position'))

		if 'Lines' in config:
			lines = config['Lines']
			if 'Highlight' in lines:
				self._view.set_highlight_current_line(
					lines.getboolean('Highlight')
				)
			if 'Numbers' in lines:
				self._view.set_show_line_numbers(lines.getboolean('Numbers'))

		if 'Wrap' in config:
			wrap = config['Wrap']
			if 'Enable' in wrap:
				wrMod = Gtk.WrapMode.NONE
				if wrap.getboolean('Enable'):
					modes = Gtk.WrapMode
					mod = modes.WORD
					if 'Anywhere' in wrap:
						mod = modes.CHAR if wrap.getboolean('Anywhere') else mod
				self._view.set_wrap_mode(mod)

		if 'Font' in config:
			font = config['Font']
			if 'Face' in font:
				face = font['Face']
				family = ' '.join(face.split(' ')[:-1])
				size = face.split(' ')[-1]

				# set font with pango
				desc = Pango.FontDescription.from_string(family)
				desc.set_absolute_size(int(size) * Pango.SCALE)
				self._view.get_pango_context().set_font_description(desc)

				# set font with css
				prov = Gtk.CssProvider.new()

				prov.load_from_data(
					str('*{font: ' + size +'px "' + family + '"}').encode()
				);
				self._view.get_style_context().add_provider(
					prov, Gtk.STYLE_PROVIDER_PRIORITY_USER
				)

		if 'Theme' in config:
			theme = config['Theme']
			manager = GtkSource.StyleSchemeManager.get_default()
			if 'Enable' in theme:
				if theme.getboolean('Enable'):
					bfr.set_highlight_syntax(True)
					if 'Id' in theme:
						style = manager.get_scheme(theme['Id'])
						assert(style)
						bfr.set_style_scheme(style)
				else:
					bfr.set_highlight_syntax(False)
					style = manager.get_scheme('mili-gtk')
					assert(style)
					bfr.set_style_scheme(style)

		if 'Keybindings' in config:
			keys = config['Keybindings']
			bindMap = {
				'Open': self._interact_open,
				'Save': self._interact_save,
				'SaveAs': self._interact_save_as,
				'New': self._interact_new,
				'Find': self._interact_find
			}
			# compare and save bindings for everything in lower case
			for purpose in bindMap.keys():
				key, mods = Gtk.accelerator_parse(keys[purpose].lower())
				key -= 32 if mods & Gdk.ModifierType.SHIFT_MASK else 0
				self._keybinds[(key, mods)] = bindMap[purpose]

	# detect and set indentation mode to match current file
	def _guess_indent(self):
		# get the content of the current file
		bfr = self._view.get_buffer()
		text = bfr.get_text(
			bfr.get_start_iter(),
			bfr.get_end_iter(),
			False
		)

		minSpaces = 0 # smallest indent is the best guess for spaces
		for line in text.split('\n'):
			# ignore whitespace-only lines
			if line.strip('\t ') == '':
				continue

			# count leading tabs and spaces
			tabs, spaces = 0, 0
			for char in line:
				tabs += 1 if char == '\t' else 0
				spaces += 1 if char == ' ' else 0
				if char not in '\t ':
					break

			# ignore mixed indent lines
			if tabs and spaces:
				break

			if tabs:
				# easy, leave width at default
				self._view.set_insert_spaces_instead_of_tabs(False)
				return
			elif spaces:
				# for spaces, look at entire file
				minSpaces = min(minSpaces, spaces) if minSpaces > 0 else spaces

		# finally, take a guess at spaces
		if minSpaces:
			self._view.set_insert_spaces_instead_of_tabs(True)
			self._view.set_indent_width(minSpaces)
			self._view.set_tab_width(minSpaces)
			return

	# force exit without any popups. to close normally, use window.close instead
	def _exit(self, _):
		self._file = ''
		self._view.get_buffer().set_modified(False)
		self._window.close()

	# for error popups
	def _popup_error(self, text: str):
		dialog = Gtk.MessageDialog(
			parent = self._window,
			destroy_with_parent = True,
			modal = True,
			buttons = Gtk.ButtonsType.OK,
			text = 'Error',
			secondary_text = text
		)
		dialog.run()
		dialog.destroy()

	# ask the user if they want to save
	def _popup_ask_save(self) -> Gtk.ResponseType:
		if not self._view.get_buffer().get_modified():
			return Gtk.ResponseType.YES

		dialog = Gtk.MessageDialog(
			parent = self._window,
			destroy_with_parent = True,
			modal = True,
			buttons = Gtk.ButtonsType.NONE,
			text = 'Save first?'
		)
		dialog.add_buttons(
			Gtk.STOCK_YES,
			Gtk.ResponseType.YES,
			Gtk.STOCK_NO,
			Gtk.ResponseType.NO,
			Gtk.STOCK_CANCEL,
			Gtk.ResponseType.CANCEL
		)

		response = dialog.run()
		if response == Gtk.ResponseType.YES:
			if self._do_safely(self._interact_save) == Gtk.ResponseType.CANCEL:
				response = Gtk.ResponseType.CANCEL

		# return value used by _exit()
		dialog.destroy()
		return response

	# read from path to self._view
	def _open(self, path):
		bfr = self._view.get_buffer()

		# reload settings - some are overridden by auto-detection
		self._configure()

		# ask to save, abort if needed
		if self._popup_ask_save() == Gtk.ResponseType.CANCEL:
			return

		# read file
		try:
			# ensure file is writeable first
			open(path, 'a+').close()
			with open(path, 'r') as file:
				bfr.begin_not_undoable_action()
				bfr.set_text(file.read())
				bfr.end_not_undoable_action()
				self._file = path
		except PermissionError:
			# closing now would prevent the new one from opening
			self._window.hide()

			# spawn new elevated window
			proc = subprocess.Popen(['pkexec', sys.argv[0], path])

			# ensure process will exit after new window is closed
			pool = concurrent.futures.ThreadPoolExecutor(max_workers = 1)
			pool.submit(proc.communicate).add_done_callback(self._exit)
			pool.shutdown(wait = False)

			# don't continue opening
			return

		# auto-detect language and set syntax highlighting
		man = GtkSource.LanguageManager.get_default()
		bfr.set_language(man.guess_language(self._file, None))

		if not bfr.get_language():
			lines = bfr.get_text(
				bfr.get_start_iter(),
				bfr.get_end_iter(),
				False
			).split('\n')

			bang = lines[0] if len(lines) > 0 else ''
			for la in sorted(man.get_language_ids(), key = len, reverse = True):
				if la in bang:
					bfr.set_language(man.get_language(la))
					break

		# match indentation type
		if self._smartIndent:
			self._guess_indent()

		# update status bar: indent
		if self._view.get_insert_spaces_instead_of_tabs():
			label = str(self._view.get_indent_width()) + ' spaces '
		else:
			label = 'tabs '
		self._barIndent.set_label(label)

		# update status bar: modified
		bfr.set_modified(False)
		self._barModified.set_label('')

		# update status bar: file
		self._barFile.set_label(' ' + self._file)

		# cursor would be at EOF by default
		bfr.place_cursor(bfr.get_start_iter())

	# file open dialog
	def _interact_open(self):
		# spawn file chooser dialog
		dialog = Gtk.FileChooserDialog(
			title = 'Open file',
			parent = self._window,
			action = Gtk.FileChooserAction.OPEN
		)
		dialog.add_buttons(
			Gtk.STOCK_CANCEL,
			Gtk.ResponseType.CANCEL,
			Gtk.STOCK_OPEN,
			Gtk.ResponseType.OK,
		)

		# get path from dialog
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			path = dialog.get_filename()
		dialog.destroy()
		if response != Gtk.ResponseType.OK:
			return

		self._open(path)

	# write from self._view to path
	def _save(self, path):
		# write to file
		with open(path, 'w') as saveFile:
			bfr = self._view.get_buffer()
			text = bfr.get_text(
				bfr.get_start_iter(),
				bfr.get_end_iter(),
				False
			)
			saveFile.write(text)
			bfr.set_modified(False)

		# match indentation type
		if self._smartIndent:
			self._guess_indent()

	# wrapper for convenience
	def _interact_save_as(self):
		self._interact_save(True)

	# file save dialog
	def _interact_save(self, prompt = False) -> Gtk.ResponseType:
		needReload = False

		# save as..
		if self._file == '' or prompt:
			# spawn file chooser dialog
			dialog = Gtk.FileChooserDialog(
				title = 'Save as',
				parent = self._window,
				action = Gtk.FileChooserAction.SAVE
			)
			dialog.add_buttons(
				Gtk.STOCK_CANCEL,
				Gtk.ResponseType.CANCEL,
				Gtk.STOCK_SAVE,
				Gtk.ResponseType.OK,
			)

			# get path from dialog
			response = dialog.run()
			if response == Gtk.ResponseType.OK:
				self._file = dialog.get_filename()
				needReload = True

			dialog.destroy()
			if response != Gtk.ResponseType.OK:
				return response

		self._save(self._file)

		# if saving a new file or saving as, reload the file after save (#23)
		if needReload:
			self._open(self._file)

		return Gtk.ResponseType.OK

	def _interact_new(self):
		if self._popup_ask_save() == Gtk.ResponseType.CANCEL:
			return

		# new file not associated with any path
		self._file = ''

		# clear buffer
		bfr = self._view.get_buffer()
		bfr.begin_not_undoable_action()
		bfr.set_text('')
		bfr.end_not_undoable_action()
		bfr.set_modified(False)

		# update status bar
		self._barModified.set_label('')
		self._barFile.set_label('')

		# reload config just in case
		self._configure()

	# toggles find/replace dropdown
	def _interact_find(self):
		bfr = self._view.get_buffer()

		self._finder.set_visible(not self._finder.get_visible())

		if self._finder.get_visible():
			# deselect, as we use selection for searching
			bfr = self._view.get_buffer()
			selection = bfr.get_selection_bounds()
			if selection:
				bfr.place_cursor(selection[0])

			# focus "find" entry
			self._finder.get_children()[0].grab_focus()
		else:
			# focus editor
			self._view.grab_focus()

	# finds and selects inputted text, case insensitive
	def _find(self, *_):
		# find/replace could use GtkSource.SearchContext instead, though for
		# this use case the current solution works fine
		bfr = self._view.get_buffer()
		scanner = bfr.get_iter_at_mark(bfr.get_insert())

		# if selection exists, skip it
		current = bfr.get_selection_bounds()
		scanner = current[1] if current else scanner

		# will need this to check if search started at beginning of file
		startOffset = bfr.get_iter_at_mark(bfr.get_insert()).get_offset()

		# find next occurence, start search at cursor
		result = scanner.forward_search(
			self._entryFind.get_text(),
			Gtk.TextSearchFlags.CASE_INSENSITIVE,
			None
		)

		# select found, if not found wrap around
		cursorMark = bfr.get_insert()
		if result:
			bfr.select_range(result[0], result[1])
		elif startOffset > 0:
			# search again (only if didn't start at beginning of file before)
			# - this prevents an infinite loop
			bfr.place_cursor(bfr.get_start_iter())
			self._find()

		# scroll so selected is visible
		self._view.scroll_to_mark(cursorMark, 0, True, 1, 0.5)

	# replaces text selected by _find, runs _find if needed
	def _replace(self, *_):
		bfr = self._view.get_buffer()

		# nothing to replace, trigger find instead
		if not bfr.get_has_selection():
			self._find()
			return

		# replace text
		newText = self._entryReplace.get_text()
		bfr.delete_selection(False, False)
		bfr.insert_at_cursor(newText, len(newText))

		# find next right away. should this be configurable?
		self._find()
